(* Definition of the CPS-transformed language *)

Require Import Coq.Lists.List.
Require Import ListFacts.
Require Import "Variable".


Section Terms.

(* term *)
Inductive cterm : Set :=
  | ct_fsvar (x : var)
  | ct_fcvar (x : var)
  | ct_unit
  | ct_sabs (x : var) (t : cterm)
  | ct_cabs (x : var) (t : cterm)
  | ct_app (tf ta : cterm).

Fixpoint fsvars_in_cterm (t : cterm) : list var :=
  match t with
  | ct_fsvar x =>
      x :: nil
  | ct_fcvar _ =>
      nil
  | ct_unit =>
      nil
  | ct_sabs x t1 =>
      remove eq_var_dec x (fsvars_in_cterm t1)
  | ct_cabs _ t1 =>
      fsvars_in_cterm t1
  | ct_app tf ta =>
      fsvars_in_cterm tf ++ fsvars_in_cterm ta
  end.

Fixpoint fcvars_in_cterm (t : cterm) : list var :=
  match t with
  | ct_fsvar _ =>
      nil
  | ct_fcvar x =>
      x :: nil
  | ct_unit =>
      nil
  | ct_sabs _ t1 =>
      (fcvars_in_cterm t1)
  | ct_cabs x t1 =>
      remove eq_var_dec x (fcvars_in_cterm t1)
  | ct_app tf ta =>
      fcvars_in_cterm tf ++ fcvars_in_cterm ta
  end.

End Terms.


Section Types.

(* type *)
Inductive ctype : Set :=
  | cy_unit
  | cy_func (TA TR : ctype)
  | cy_bottom.

End Types.


Section Contexts.

(* typing context *)
Definition cctxt : Set :=
  list (var * ctype).

Definition domain_of_cctxt : cctxt -> list var :=
  map (@fst _ _).

End Contexts.


Section Rules.

(* term typing rule *)
Inductive cterm_typing : cctxt -> cctxt -> cterm -> ctype -> Prop :=
  | ctt_fsvar :
      forall (Cs Cc : cctxt) (x : var) (T : ctype),
      in' x T Cs ->
      cterm_typing Cs Cc (ct_fsvar x) T
  | ctt_fcvar :
      forall (Cs Cc : cctxt) (x : var) (T : ctype),
      in' x T Cc ->
      cterm_typing Cs Cc (ct_fcvar x) T
  | ctt_unit :
      forall Cs Cc : cctxt,
      cterm_typing Cs Cc ct_unit cy_unit
  | ctt_sabs :
      forall (Cs Cc : cctxt) (x : var) (TA TR : ctype) (t : cterm),
      cterm_typing ((x, TA) :: Cs) Cc t TR ->
      cterm_typing Cs Cc (ct_sabs x t) (cy_func TA TR)
  | ctt_cabs :
      forall (Cs Cc : cctxt) (x : var) (TA TR : ctype) (t : cterm),
      cterm_typing Cs ((x, TA) :: Cc) t TR ->
      cterm_typing Cs Cc (ct_cabs x t) (cy_func TA TR)
  | ctt_app :
      forall (Cs Cc : cctxt) (TA TR : ctype) (tf ta : cterm),
      cterm_typing Cs Cc tf (cy_func TA TR) ->
      cterm_typing Cs Cc ta TA ->
      cterm_typing Cs Cc (ct_app tf ta) TR.

End Rules.


Bind Scope ctype_scope with ctype.
Delimit Scope ctype_scope with ctype.
Notation "! T" := (cy_func T cy_bottom)
  (at level 35) : ctype_scope.
